#!/usr/bin/env python

import sys
import signal
import os

from sys import exit, argv
from select import select
from getopt import getopt
from string import atoi, split, strip, lower, find
from pickle import load
from socket import socket, gethostbyname, SOCK_STREAM, AF_INET, gethostname, error
from estutil import true, false, striplist, unpackdict, dict_ske, dict_pkr, dict_pkra, packdict, dict_port, localhost, sigchld_handler, ctrlc_handler, keypair
from os import fork, execv
from time import time, strftime, localtime

try:
    from syslog import syslog
    syslog_support = true
except:
    print "Warning: your version of python doesn't have syslog support compiled in !"
    syslog_support = false

from Crypto.Cipher import ARC4
from CryptoWorld.PublicKey import RSA

class Parameter:

    def __init__(self):
	self.port = None
	self.keyfile = None
	self.maxconn = 16
	self.verbose = false
	self.accesstab = None
	self.syslog = false

    def usage(self):
	print "Usage: estd.py -p <port> -f <keyfile> -m <maxconn> -a <accesstable> -s -v"
	exit()

    def parse(self):
	try:
	    optlist, args = getopt(argv[1:], "p:f:m:a:sv")
	except:
	    self.usage()

	for flag, arg in optlist:
	    if (flag == "-p"): self.port = atoi(arg)
	    elif (flag == "-f"): self.keyfile = arg
	    elif (flag == "-m"): self.maxconn = atoi(arg)
	    elif (flag == "-v"): self.verbose = true
	    elif (flag == "-a"): self.accesstab = arg
	    elif (flag == "-s") and (syslog_support == true): self.syslog = true

	if (len(args) > 0) or (self.port == None) or (self.keyfile == None) or (self.accesstab == None):
	    self.usage()

class Server:

    def __init__(self):
	self.socket = None
	self.keypair = None
	self.publickey = None
	self.maxconn = 16
	self.port = None
	self.daemoncrypto = {}                             # Cipher for stuff the daemon encrypts
	self.clientcrypto = {}                             # Cipher for encrypting things from the client
	self.verbose = false
	self.accesstab = {}
	self.syslog = None

    def log(self, message):
	if (self.syslog == true):
	    syslog("estd.py: %s" % message)
	if (self.verbose == true):
	    sys.stdout.write("%s estd.py: %s\n" % (strftime("%b %d %X", localtime(time())), message))

    def initialize(self, keyfile, maxconn, port, verbose, accesstab, syslog):

	self.verbose = verbose
	self.syslog = syslog

	self.log("Initializing ECLiPt Secure Tunnel Daemon")

	if not os.path.exists(keyfile):
	    self.log("Key file %s doesn't exist" % keyfile)
	    exit(1)

	file = open(keyfile, "r")
	estpair = load(file)
	file.close()

	self.keypair = RSA()
	self.keypair.set_public_key(estpair.public_key)
	self.keypair.set_private_key(estpair.private_key)

	self.publickey = estpair.public_key
	self.maxconn = maxconn
	self.port = port

	if not os.path.exists(accesstab):
	    self.log("Access table file %s doesn't exist" % accesstab)
	    exit(1)

	file = open(accesstab, "r")
	table = striplist(file.readlines())
	file.close()

	self.log("Reading local access table")

	for i in range(len(table)):
	    if (table[i][0] != "#") and (table[i] != ""):
		splitted = split(table[i], ":")
		if (len(splitted) != 2):
		    self.log("Invalid access table entry in %s line %d" % (accesstab, i))
		else:
		    port, hosts = splitted
		    port = atoi(strip(port))
		    hostlist = striplist(split(hosts, ","))
		    self.accesstab[port] = hostlist

	self.log("Access table successfully read")

    def listen(self):

	self.socket = socket(AF_INET, SOCK_STREAM)
	try:
	    self.socket.bind (gethostbyname(gethostname()), self.port)
	except error, detail:
	    self.log("Couldn't bind to port %d: %s" % (self.port, detail[1]))
	    exit(1)
	self.socket.listen (self.maxconn)

	self.log("Bind to port %d successful" % self.port)

	while 1:
	    try:
		[iwtd, owtd, ewtd] = select ([self.socket], [], [])
	    except:
		continue

	    for in_sock in iwtd:
		in_conn, in_addr = in_sock.accept()
		connection_ok = true

		try:
		    length = atoi(in_conn.recv(4))            # Receive 4 Byte (session key packet length)
		    packet = unpackdict(in_conn.recv(length)) # Receive packet and unpack it
		except (ValueError, error), detail:
		    self.log("[ host %s ] Unauthorized access !" % in_addr[0])
		    in_conn.close()
		    connection_ok = false

		if (connection_ok == true):

		    if packet.has_key(dict_ske):
			self.log("[ host %s ] Getting session key" % in_addr[0])
			sessionkey = self.keypair.private_decrypt(packet[dict_ske])
			self.daemoncrypto[in_addr[0]] = ARC4.new(sessionkey)
			self.clientcrypto[in_addr[0]] = ARC4.new(sessionkey)
			self.log("[ host %s ] Got session key" % in_addr[0])
		    elif packet.has_key(dict_pkr):
			self.log("[ host %s ] Got public key request" % in_addr[0])
			in_conn.send(packdict({dict_pkra:self.publickey}))
			self.log("[ host %s ] Public key sent" % in_addr[0])
		    elif packet.has_key(dict_port):
			destport = packet[dict_port]

			self.log("[ host %s ] Connection requested" % in_addr[0])

			allowed = false

			if self.accesstab.has_key(destport):
			    hostlist = self.accesstab[destport]
			    if (lower(hostlist[0]) == "all"):
				allowed = true
			    else:
				for i in range(len(hostlist)):
				    if (find(in_addr[0], hostlist[i]) != -1):
					allowed = true

			if (allowed == false):

			    self.log("[ host %s -> %d ] Forwarding refused !" % (in_addr[0], destport))
			    in_conn.close()

			else:

			    self.log("[ host %s -> %d ] Forwarding requests" % (in_addr[0], destport))

			    if (fork() == 0):
				out_sock = socket(AF_INET, SOCK_STREAM)
				out_sock.connect(localhost, destport)

				if not self.daemoncrypto.has_key(in_addr[0]):
				    self.log("[ host %s -> %d ] Connection closed: no session key available" % (in_addr[0], destport))
				    in_conn.close()
				    out_sock.close()
				    execv('/bin/true', ())

				daemoncrypto = self.daemoncrypto[in_addr[0]]
				clientcrypto = self.clientcrypto[in_addr[0]]

				self.log("[ host %s -> %d ] Connection established" % (in_addr[0], destport))
				data = ''
				while 1:
				    [iwtd, owtd, ewtd] = select([in_conn, out_sock],[],[])
				    if in_conn in iwtd:
					try:
					    data = in_conn.recv(1024)
					    if not data: break
					    plain = clientcrypto.decrypt(data)
					    out_sock.send(plain)
					except error, detail:
					    self.log("[ host %s -> %d ] Connection lost: %s" % (in_addr[0], destport, detail[1]))
					    exit(1)
				    if out_sock in iwtd:
					try:
					    data = out_sock.recv(1024)
					    if not data: break
					    cipher = daemoncrypto.encrypt(data)
					    in_conn.send(cipher)
					except error, detail:
					    self.log("[ host %s -> %d ] Connection lost: %s" % (in_addr[0], destport, detail[1]))
					    exit(1)
				self.log("[ host %s -> %d ] Connection closed" % (in_addr[0], destport))
				in_conn.close()
				out_sock.close()
				execv('/bin/true', ())
			    else:
				in_conn.close()

parameter = Parameter()
parameter.parse()

signal.signal(signal.SIGINT, ctrlc_handler)
signal.signal(signal.SIGCHLD, sigchld_handler)

server = Server()
server.initialize(parameter.keyfile, parameter.maxconn, parameter.port, parameter.verbose, parameter.accesstab, parameter.syslog)
server.listen()
